from io import BufferedReader
from pathlib import Path
from typing import Dict, List, Optional

from pydantic import (
    UUID4,
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    SecretStr,
    StringConstraints,
    field_serializer,
    field_validator,
    model_validator,
)
from typing_extensions import Annotated

from promptquality.constants.dataset_format import DatasetFormat
from promptquality.constants.integrations import IntegrationName
from promptquality.constants.job import JobStatus
from promptquality.constants.run import RunDefaults, TagType
from promptquality.constants.scorers import Scorers
from promptquality.types.prompt_optimization import PromptOptimizationConfiguration
from promptquality.types.registered_scorers import RegisteredScorer
from promptquality.types.settings import Settings
from promptquality.utils.dataset import parse_dataset
from promptquality.utils.name import ts_name


class CreateProjectRequest(BaseModel):
    name: str
    type: str = RunDefaults.project_type

    @field_validator("name", mode="before")
    def generate_name(cls, value: Optional[str]) -> str:
        if not value:
            value = ts_name(prefix="project")
        return value


class TemplateVersion(BaseModel):
    name: str = Field(description="The name of the template")
    version: Optional[int] = Field(default=None, description="The template version, defaults to the production version")


class BaseTemplateVersionRequest(BaseModel):
    template: str
    version: Optional[int] = None


class CreateTemplateRequest(BaseTemplateVersionRequest):
    name: str
    project_id: UUID4

    @field_validator("name", mode="before")
    def generate_name(cls, value: Optional[str]) -> str:
        if not value:
            value = ts_name(prefix="template")
        return value


class CreateTemplateVersionRequest(BaseTemplateVersionRequest):
    template_id: UUID4
    project_id: UUID4


class BaseTemplateVersionResponse(BaseTemplateVersionRequest):
    id: UUID4


class CreateTemplateVersionResponse(BaseTemplateVersionResponse):
    version: int


class BaseTemplateResponse(BaseModel):
    id: UUID4
    name: str
    template: str
    selected_version: CreateTemplateVersionResponse
    selected_version_id: UUID4
    all_versions: List[CreateTemplateVersionResponse] = Field(default_factory=list)


class BaseDatasetRequest(BaseModel):
    format: DatasetFormat = DatasetFormat.csv
    file_path: Path = Field(exclude=True)

    @model_validator(mode="before")
    def dataset_to_path(cls, values: Dict) -> Dict:
        if "file_path" in values:
            values["file_path"], values["format"] = parse_dataset(values["file_path"])
        return values

    @property
    def files(self) -> Dict[str, BufferedReader]:
        return dict(file=self.file_path.open("rb"))


class UploadDatasetRequest(BaseDatasetRequest):
    project_id: UUID4
    prompt_template_version_id: UUID4
    file_path: Path = Field(exclude=True)

    @property
    def params(self) -> Dict[str, str]:
        return dict(
            prompt_template_version_id=str(self.prompt_template_version_id),
            format=self.format.value,
        )


class Dataset(BaseModel):
    id: UUID4 = Field(validation_alias=AliasChoices("dataset_id", "id"))
    file_name: Optional[str] = None
    num_rows: Optional[int] = None


class EstimateCostRequest(BaseDatasetRequest, Settings):
    project_id: UUID4 = Field(exclude=True)
    template: str = Field(serialization_alias="prompt_template")

    @property
    def params(self) -> Dict[str, str]:
        return self.model_dump(mode="json", by_alias=True, exclude_unset=True, exclude_none=True)


class EstimatedCostResponse(BaseModel):
    total_cost: float
    query_cost: float
    log_probs_cost: float


class RunTag(BaseModel):
    key: Annotated[str, StringConstraints(max_length=256)]
    value: Annotated[str, StringConstraints(max_length=256)]
    tag_type: TagType


class CreateRunRequest(BaseModel):
    name: str
    project_id: UUID4
    task_type: int = RunDefaults.prompt_evaluation_task_type
    run_tags: List[RunTag] = Field(default_factory=list)

    @field_validator("name", mode="before")
    def generate_name(cls, value: Optional[str]) -> str:
        if not value:
            value = ts_name(prefix="run")
        return value


class RunResponse(CreateRunRequest):
    id: UUID4


class ProjectResponse(CreateProjectRequest):
    id: UUID4
    runs: List[RunResponse] = Field(default_factory=list)


class ScorerSettings(BaseModel):
    scorer_name: str


class ScorersConfiguration(BaseModel):
    """
    Configuration to control which scorers to enable and disable.

    Can be used in runs and chain runs, with or instead of scorers arg. scorers explicitly set in scorers arg will
    override this.
    """

    adherence_nli: bool = False
    chunk_attribution_utilization_gpt: bool = False
    chunk_attribution_utilization_nli: bool = False
    completeness_gpt: bool = False
    completeness_nli: bool = False
    context_relevance: bool = False
    factuality: bool = False
    groundedness: bool = False
    pii: bool = True
    prompt_injection: bool = False
    prompt_perplexity: bool = False
    sexist: bool = False
    tone: bool = False
    toxicity: bool = True
    uncertainty: bool = False

    @model_validator(mode="after")
    def disallow_conflicts(self) -> "ScorersConfiguration":
        """Raise Value Error if conflicting scorers are selected."""
        source = None
        if self.adherence_nli and self.groundedness:
            source = "adherence"
        elif self.chunk_attribution_utilization_gpt and self.chunk_attribution_utilization_nli:
            source = "attribution/utilization"
        elif self.completeness_gpt and self.completeness_nli:
            source = "completeness"
        if source:
            raise ValueError(f"Cannot use basic {source} with plus {source}")
        return self

    @classmethod
    def from_scorers(cls, scorers: List[Scorers]) -> "ScorersConfiguration":
        return cls(**{scorer.value: True for scorer in scorers})

    def merge_scorers(self, scorers: List[Scorers]) -> "ScorersConfiguration":
        return self.model_copy(update={scorer.value: True for scorer in scorers})


class CreateJobRequest(BaseModel):
    project_id: UUID4
    run_id: UUID4
    prompt_dataset_id: Optional[UUID4] = None
    prompt_template_version_id: Optional[UUID4] = None
    prompt_settings: Optional[Settings] = None
    prompt_scorers_configuration: Optional[ScorersConfiguration] = None
    prompt_scorer_settings: Optional[ScorerSettings] = None
    prompt_registered_scorers_configuration: Optional[List[RegisteredScorer]] = None
    job_name: str = RunDefaults.prompt_run_job_name
    task_type: Optional[int] = RunDefaults.prompt_evaluation_task_type
    prompt_optimization_configuration: Optional[PromptOptimizationConfiguration] = None


class JobInfoMixin(BaseModel):
    job_id: UUID4
    link: str
    job_name: Optional[str] = None


class CreateJobResponse(CreateJobRequest, JobInfoMixin):
    ...


class GetMetricsRequest(BaseModel):
    project_id: UUID4
    run_id: UUID4


class PromptMetrics(BaseModel):
    total_responses: Optional[int] = None
    average_hallucination: Optional[float] = None
    average_bleu: Optional[float] = None
    average_rouge: Optional[float] = None
    average_cost: Optional[float] = None
    total_cost: Optional[float] = None

    model_config = ConfigDict(extra="allow")


class GetJobStatusResponse(BaseModel):
    id: UUID4
    project_id: UUID4
    run_id: UUID4
    status: JobStatus
    job_name: str = RunDefaults.prompt_run_job_name
    error_message: Optional[str] = None
    progress_message: Optional[str] = None
    steps_completed: int = 0
    steps_total: int = 0
    progress_percent: float = 0.0
    request_data: Optional[CreateJobRequest] = None


class CreateIntegrationRequest(BaseModel):
    api_key: SecretStr = Field(serialization_alias="token")
    name: str = Field(exclude=True)
    organization_id: Optional[str] = None
    endpoint: Optional[str] = None
    headers: Optional[Dict[str, str]] = None
    proxy: Optional[bool] = None

    @field_validator("name", mode="before")
    def validate_name(cls, value: Optional[str]) -> str:
        if value not in vars(IntegrationName):
            raise ValueError(f"Integration name must be one of {vars(IntegrationName)}.")
        return value

    @field_serializer("api_key", when_used="always")
    def serialize_api_key(self, value: SecretStr) -> str:
        return value.get_secret_value()


class SelectTemplateVersionRequest(BaseModel):
    project_id: UUID4
    template_id: UUID4
    version: int


class UserSubmittedMetricsResponse(BaseModel):
    project_id: UUID4
    run_id: UUID4
    job_id: UUID4
    scorer_name: str



from pathlib import Path
from typing import List, Optional, Union

from promptquality.set_config import set_config
from promptquality.types.config import Config
from promptquality.types.registered_scorers import (
    ListRegisteredScorers,
    RegisteredScorer,
)
from promptquality.utils.logger import logger
from promptquality.utils.name import check_scorer_name


def register_scorer(
    scorer_name: str, scorer_file: Union[str, Path], config: Optional[Config] = None
) -> RegisteredScorer:
    config = config or set_config()
    scorer_file = Path(scorer_file)
    if not scorer_file.exists():
        raise FileNotFoundError(f"Scorer file {scorer_file} does not exist.")
    check_scorer_name(scorer_name)
    response_dict = config.api_client.register_scorer(scorer_name, scorer_file)

    score_type = response_dict.get("score_type")
    score_type = score_type if score_type else "None"

    scoreable_node_types = response_dict.get("scoreable_node_types")
    scoreable_node_types = ", ".join(scoreable_node_types) if scoreable_node_types else "None"

    print(
        f"🔢 Registered scorer '{scorer_name}'. Score type: {score_type}, scorable node types: {scoreable_node_types}."
    )
    scorer = RegisteredScorer.model_validate(response_dict)
    logger.debug(f"Registered scorer {scorer.name} with id {scorer.id}.")
    return scorer


def list_registered_scorers(config: Optional[Config] = None) -> List[RegisteredScorer]:
    config = config or set_config()
    response_dict = config.api_client.list_registered_scorers()
    scorers = ListRegisteredScorers.model_validate(response_dict)
    logger.debug(f"Found {len(scorers.scorers)} registered scorers.")
    return scorers.scorers



from types import SimpleNamespace

from promptquality.constants.run import RunDefaults

DEFAULT_TEMPERATURE = 1
DEFAULT_MAX_TOKENS = 4096
PromptOptimizationDefaults = SimpleNamespace(
    project_type=RunDefaults.project_type,
    prompt_optimization_job_name="prompt_optimization",
    prompt_evaluation_task_type=RunDefaults.prompt_evaluation_task_type,
)



from collections import ChainMap
from csv import DictReader, DictWriter
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict, List, Optional, Tuple, Union
from warnings import warn

from promptquality.constants.dataset_format import DatasetFormat
from promptquality.utils.dependencies import is_pyarrow_available

DatasetType = Union[Path, str, Dict[str, List[str]], List[Dict[str, str]]]


def transpose_dict_to_rows(dataset: Dict[str, List[str]]) -> List[Dict[str, str]]:
    key_rows = [[{key: value} for value in values] for key, values in dataset.items()]
    return [dict(ChainMap(*key_row)) for key_row in zip(*key_rows)]


def parse_dataset(dataset: DatasetType) -> Tuple[Path, DatasetFormat]:
    """
    Set the dataset path and format.

    If the dataset is a dictionary or list, it will be stored as a a temporary file and
    the path to the temporary file will be returned. If the dataset is a path or string,
    the path will be returned. If the dataset is not a dictionary, list, path, or
    string, a ValueError will be raised.

    If the dataset is a dictionary or list and PyArrow is available, the dataset will
    be stored as a Feather file. Otherwise, the dataset will be stored as a CSV file.

    Parameters
    ----------
    dataset : DatasetType
        Dataset as provided by the user.

    Returns
    -------
    Tuple[Path, DatasetFormat]
        Path to the dataset file and the dataset format.

    Raises
    ------
    ValueError
        If the dataset is not a dictionary, list, path, or string.
    ValueError
        If the dataset file is not a CSV or Feather file.
    FileNotFoundError
        If the dataset file does not exist.
    """
    # Set the dataset path.
    dataset_path: Optional[Path] = None
    if isinstance(dataset, (dict, list)):
        dataset_rows = transpose_dict_to_rows(dataset) if isinstance(dataset, dict) else dataset
        if is_pyarrow_available:
            with NamedTemporaryFile(mode="wb", suffix=f".{DatasetFormat.feather}", delete=False) as dataset_file:
                from pyarrow import Table  # type: ignore[import]
                from pyarrow.feather import write_feather  # type: ignore[import]

                write_feather(Table.from_pylist(dataset_rows), dataset_file.name)
                dataset_path = Path(dataset_file.name)
        else:
            with NamedTemporaryFile(mode="wt", suffix=f".{DatasetFormat.csv}", delete=False) as dataset_file:
                field_names = dataset_rows[0].keys()
                writer = DictWriter(dataset_file, fieldnames=field_names)
                writer.writeheader()
                writer.writerows(dataset_rows)
                dataset_path = Path(dataset_file.name)
    elif isinstance(dataset, Path):
        dataset_path = dataset
    elif isinstance(dataset, str):
        dataset_path = Path(dataset)
    else:
        raise ValueError(f"Invalid dataset type: {type(dataset)}.")
    if not dataset_path.exists():
        raise FileNotFoundError(f"Dataset file {dataset_path} does not exist.")

    # Set the dataset format.
    if dataset_path.suffix.endswith(DatasetFormat.csv):
        dataset_format = DatasetFormat.csv
        # Check if we have embedding columns.
        reader = DictReader(dataset_path.open("rt", errors="ignore"))
        for column in reader.fieldnames or []:
            if column.endswith("_embedding"):
                warn(
                    f"Column {column} appears to be an embedding column, please use "
                    "a feather file as dataset format instead."
                )
    elif dataset_path.suffix.endswith(DatasetFormat.jsonl):
        dataset_format = DatasetFormat.jsonl
    elif dataset_path.suffix.endswith(DatasetFormat.feather):
        dataset_format = DatasetFormat.feather
    else:
        raise ValueError(f"Dataset file {dataset_path} must be a CSV or Feather file.")

    return (dataset_path, dataset_format)



from pathlib import Path
from typing import List, Optional
from warnings import warn

from pydantic import UUID4

from promptquality.helpers import (
    create_project,
    create_prompt_optimization_job,
    create_run,
    create_template,
    get_job_status,
    upload_dataset,
)
from promptquality.set_config import set_config
from promptquality.types.config import Config
from promptquality.types.prompt_optimization import (
    PromptOptimizationConfiguration,
    PromptOptimizationEpochResult,
    PromptOptimizationResults,
)
from promptquality.types.run import GetMetricsRequest
from promptquality.utils.name import ts_name

PROMPT_OPTIMIZATION_WARNING = (
    "The optimize_prompt feature is not yet release for public use. "
    "Please be aware that the response will not be useful until the "
    "feature is fully released."
)


def optimize_prompt(
    prompt_optimization_config: PromptOptimizationConfiguration,
    dataset: Path,
    project_name: Optional[str] = None,
    run_name: Optional[str] = None,
    config: Optional[Config] = None,
) -> UUID4:
    """
    Optimize a prompt for a given task.

    This function takes a prompt and a list of evaluation criteria, and optimizes the
    prompt for the given task. The function uses the OpenAI API to generate and evaluate
    prompts, and returns the best prompt based on the evaluation criteria.

    Parameters
    ----------
    prompt_optimization_config : PromptOptimizationConfiguration
        Configuration for the prompt optimization job.
    dataset : Path
        Path to the training dataset.
    project_name : Optional[str], optional
        Name of the project, by default None. If None we will generate a name.
    run_name : Optional[str], optional
        Name of the run, by default None. If None we will generate a name.
    config : Optional[Config], optional
        pq config object, by default None. If None we will use the default config.

    Returns
    -------
    job_id: UUID4
        Unique identifier required to fetch Prompt Optimization results.
    """
    warn(PROMPT_OPTIMIZATION_WARNING)
    config = config or set_config()

    project = create_project(project_name, config)
    template_response = create_template(
        prompt_optimization_config.prompt,
        project.id,
        # Use project name as template name if not provided.
        template_name=project.name,
        config=config,
    )
    dataset_id = upload_dataset(
        dataset,
        project.id,
        template_response.selected_version_id,
        config,
    )
    run_id = create_run(
        project.id,
        run_name=run_name or ts_name(prefix=f"{template_response.name}-v{template_response.selected_version.version}"),
        config=config,
    )

    job_id = create_prompt_optimization_job(
        prompt_optimization_configuration=prompt_optimization_config,
        project_id=project.id,
        run_id=run_id,
        train_dataset_id=dataset_id,
        config=config,
    )

    return job_id


def fetch_prompt_optimization_result(
    job_id: Optional[UUID4] = None,
    config: Optional[Config] = None,
) -> PromptOptimizationResults:
    """
    Fetch the prompt optimization results.

    Parameters
    ----------
    job_id : UUID4
        Unique identifier required to fetch Prompt Optimization results.

    Returns
    -------
    PromptOptimizationResults
        - best_prompt: The best prompt based on the evaluation criteria.
        - train_results: List of epoch results for the training dataset.
            Sorted by epoch ascending.
        - val_results: List of epoch results for the validation dataset.
            Sorted by epoch ascending.
    """
    warn(PROMPT_OPTIMIZATION_WARNING)
    config = config or set_config()
    job_id = job_id or config.current_prompt_optimization_job_id
    if job_id is None:
        raise ValueError("job_id is required.")

    job = get_job_status(job_id, config)
    project_id, run_id = job.project_id, job.run_id

    results: List[PromptOptimizationEpochResult] = []
    metrics_request = GetMetricsRequest(project_id=project_id, run_id=run_id)
    all_metrics = config.api_client.get_metrics(metrics_request)
    for metric in all_metrics:
        results.append(
            PromptOptimizationEpochResult(
                key=metric["key"],
                epoch=metric["epoch"],
                **metric["extra"],
            )
        )

    results = [result for result in results]
    best_prompt = max(results, key=lambda x: x.rating).new_prompt
    return PromptOptimizationResults(best_prompt=best_prompt, results=results)



from typing import Callable, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator

from promptquality.types.rows import PromptRow
from promptquality.utils.name import check_scorer_name

CustomMetricType = Union[float, int, bool, str, None]


class CustomScorer(BaseModel):
    name: str
    scorer_fn: Callable[[PromptRow], CustomMetricType] = Field(validation_alias="executor")
    aggregator_fn: Optional[Callable[[List[CustomMetricType], List[int]], Dict[str, CustomMetricType]]] = Field(
        default=None, validation_alias="aggregator"
    )

    model_config = ConfigDict(populate_by_name=True)

    @field_validator("name", mode="before")
    def validate_scorer_name(cls, name: str) -> str:
        return check_scorer_name(name)



from itertools import product
from typing import Callable, Dict, Iterable

from tqdm.auto import tqdm


def sweep(fn: Callable, params: Dict[str, Iterable]) -> None:
    """
    Run a sweep of a function over various settings.

    Given a function and a dictionary of parameters, run the function over all
    combinations of the parameters.

    Parameters
    ----------
    fn : Callable
        Function to run.

    params : Dict[str, Iterable]
        Dictionary of parameters to run the function over. The keys are the
        parameter names and the values are the values to run the function with.
    """
    all_combinations = list(product(*params.values()))
    for combination in tqdm(all_combinations):
        fn(**dict(zip(params.keys(), combination)))



from getpass import getpass
from pathlib import Path
from posixpath import join
from time import time
from typing import Optional
from urllib.parse import urljoin
from webbrowser import open_new_tab

from jwt import decode
from pydantic import (
    UUID4,
    HttpUrl,
    SecretStr,
    ValidationError,
    computed_field,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic_settings import BaseSettings, SettingsConfigDict

from promptquality.constants.prompt_optimization import PromptOptimizationDefaults
from promptquality.constants.routes import Routes
from promptquality.constants.run import RunDefaults
from promptquality.types.run import (
    BaseTemplateResponse,
    CreateTemplateVersionResponse,
    Dataset,
    JobInfoMixin,
    ProjectResponse,
    RunResponse,
)
from promptquality.utils.api_client import ApiClient
from promptquality.utils.config import get_config_location
from promptquality.utils.logger import logger


class Config(BaseSettings):
    console_url: HttpUrl
    # User.
    username: Optional[str] = None
    password: Optional[SecretStr] = None
    api_key: Optional[SecretStr] = None
    token: Optional[SecretStr] = None
    current_user: Optional[str] = None
    # Project.
    current_project_id: Optional[UUID4] = None
    current_project_name: Optional[str] = None
    # Run.
    current_run_id: Optional[UUID4] = None
    current_run_name: Optional[str] = None
    current_run_url: Optional[str] = None
    current_run_task_type: Optional[int] = None
    # Template.
    current_template_id: Optional[UUID4] = None
    current_template_name: Optional[str] = None
    # Version
    current_template_version_id: Optional[UUID4] = None
    current_template_version: Optional[int] = None
    current_template: Optional[str] = None
    # Dataset.
    current_dataset_id: Optional[UUID4] = None
    # Job.
    current_job_id: Optional[UUID4] = None
    current_prompt_optimization_job_id: Optional[UUID4] = None

    model_config = SettingsConfigDict(
        # Allow loading from environment variables.
        env_prefix="GALILEO_",
        # Ignore unknown fields when loading from a config file.
        extra="ignore",
        # Validate fields at assignment.
        validate_assignment=True,
    )

    @computed_field
    def api_url(self) -> HttpUrl:
        # Local dev.
        if self.console_url.host in ["localhost", "127.0.0.1"]:
            api_url = "http://localhost:8088"
        else:
            api_url = self.console_url.unicode_string().replace("console", "api")
        return HttpUrl(api_url)

    @property
    def config_file(self) -> Path:
        return get_config_location()

    @field_validator("console_url", mode="before")
    def http_url(cls, value: str) -> str:
        if value and not (value.startswith("https") or value.startswith("http")):
            value = f"https://{value}"
        return value

    @model_validator(mode="after")
    def validate_api_url(self) -> "Config":
        if not ApiClient.healthcheck(str(self.api_url)):
            # TODO: Make this a custom error.
            raise ValidationError(f"Could not connect to {self.api_url}.")
        return self

    @field_serializer("token", when_used="json")
    def serialize_token(self, value: SecretStr) -> str:
        return value.get_secret_value()

    @classmethod
    def read(cls) -> "Config":
        return cls.model_validate_json(get_config_location().read_text())

    def write(self) -> None:
        self.config_file.parent.mkdir(exist_ok=True)
        self.config_file.write_text(self.model_dump_json(exclude_none=True))

    @property
    def project_url(self) -> str:
        if not self.current_project_id:
            raise ValueError("No project set.")
        return join(
            self.console_url.unicode_string(),
            "prompt",
            "chains" if self.current_run_task_type == RunDefaults.prompt_chain_task_type else "",
            f"{self.current_project_id}",
        )

    @property
    def api_client(self) -> ApiClient:
        self._refresh_token()
        if not self.token:
            raise ValueError("No token set. Please log in.")
        return ApiClient(api_url=self.api_url, token=self.token)

    def token_login(self) -> str:
        token_url = urljoin(str(self.console_url), Routes.get_token)
        logger.info(f"🔐 Opening {token_url} to generate a new Galileo token.")
        try:
            open_new_tab(token_url)
        except Exception:
            pass
        finally:
            print(f"Go to {token_url} to generate a new Galileo token.")
            return getpass("🔐 Enter your Galileo token:")

    def login(self) -> None:
        if self.api_key:
            token = ApiClient.api_key_login(str(self.api_url), self.api_key.get_secret_value()).get("access_token", "")
        elif self.username and self.password:
            token = ApiClient.username_login(str(self.api_url), self.username, self.password.get_secret_value()).get(
                "access_token", ""
            )
        else:
            token = self.token_login()
        self.token = SecretStr(token)
        self.current_user = self.api_client.get_current_user().get("email")
        self.write()
        print(f"👋 You have logged into 🔭 Galileo ({self.console_url}) as {self.current_user}.")

    def _refresh_token(self) -> None:
        """Refresh token if not present or expired."""
        # Check to see if our token is expired before making a request
        # and refresh token if it's expired
        if self.token:
            claims = decode(self.token.get_secret_value(), options={"verify_signature": False})
            if claims.get("exp", 0) < time():
                self.login()

    def logout(self) -> None:
        # Reset credentials.
        self.username = None
        self.password = None
        self.token = None
        self.current_user = None
        # Reset all other values.
        self.current_project_id = None
        self.current_project_name = None
        self.current_run_id = None
        self.current_run_name = None
        self.current_run_task_type = None
        self.current_template_id = None
        self.current_template_name = None
        self.current_template_version_id = None
        self.current_template_version = None
        self.current_template = None
        self.current_dataset_id = None
        self.current_job_id = None
        self.current_prompt_optimization_job_id = None
        self.write()
        print("👋 You have logged out of 🔭 Galileo.")

    def merge_project(self, project: ProjectResponse) -> None:
        self.current_project_id = project.id
        self.current_project_name = project.name
        self.write()
        logger.debug(f"📝 Set current project to {project.name}.")

    def merge_template(self, template: BaseTemplateResponse) -> None:
        self.current_template_id = template.id
        self.current_template_name = template.name
        self.merge_template_version(template.selected_version)
        self.write()
        logger.debug(f"📝 Set current template to {template.name}.")

    def merge_template_version(self, template_version: CreateTemplateVersionResponse) -> None:
        self.current_template_version_id = template_version.id
        self.current_template_version = template_version.version
        self.current_template = template_version.template
        self.write()
        logger.debug(f"📝 Set current template version to {template_version.version}.")

    def merge_dataset(self, dataset: Dataset) -> None:
        self.current_dataset_id = dataset.id
        self.write()
        logger.debug(f"📝 Set current dataset to {dataset.id}.")

    def merge_run(self, run: RunResponse) -> None:
        self.current_run_id = run.id
        self.current_run_name = run.name
        self.current_run_task_type = run.task_type
        self.write()
        logger.debug(f"📝 Set current run to {run.name}.")

    def merge_job(self, job_info: JobInfoMixin) -> None:
        self.current_job_id = job_info.job_id
        self.current_run_url = job_info.link
        if job_info.job_name == PromptOptimizationDefaults.prompt_optimization_job_name:
            self.current_prompt_optimization_job_id = job_info.job_id
        self.write()
        logger.debug(f"📝 Set current job to {job_info.job_id}.")



from enum import Enum
from http.client import HTTPException
from typing import Any, Callable, Dict, Optional, Union
from urllib.parse import urljoin

from requests import Response


def _validate_response(response: Response) -> Dict[str, Any]:
    if response.ok:
        return response.json()

    error_message = response.text
    try:
        if detail := response.json().get("detail"):
            error_message = detail
    finally:
        error_message = f"❌ Something didn't go quite right. Error message: {error_message}"
        raise HTTPException(error_message) from None


def make_request(
    request_method: Callable,
    base_url: str,
    endpoint: str,
    body: Optional[Dict] = None,
    data: Optional[Dict] = None,
    files: Optional[Dict] = None,
    params: Optional[Dict] = None,
    headers: Optional[Dict[str, str]] = None,
    timeout: Union[int, None] = None,
) -> Any:
    response = request_method(
        urljoin(base_url, endpoint),
        json=body,
        data=data,
        files=files,
        params=params,
        headers=headers or {},
        timeout=timeout,
    )
    _validate_response(response)
    return response.json()


class HttpHeaders(str, Enum):
    accept = "accept"
    content_type = "Content-Type"
    application_json = "application/json"

    @staticmethod
    def accept_json() -> Dict[str, str]:
        return {HttpHeaders.accept: HttpHeaders.application_json}

    @staticmethod
    def content_type_json() -> Dict[str, str]:
        return {HttpHeaders.content_type: HttpHeaders.application_json}

    @staticmethod
    def json() -> Dict[str, str]:
        return {**HttpHeaders.accept_json(), **HttpHeaders.content_type_json()}
